package imp.ringing;

import android.util.Log;

import java.util.Timer;
import java.util.TimerTask;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;

/**
 * A protocol that encodes data into the length of ringing. 
 */
public class LengthProtocol extends Protocol {
	private Queue<Boolean> queue;
	private final int BLOCK_SIZE;
	private final long GRANULARITY_MS;
	private long elapsedTime;
	
	/**
	 * Initializes the Length protocol and finds the optimal block size.
	 * @param params Parameters.
	 */
	public LengthProtocol(Parametrization params) {
		super(params);
		
		GRANULARITY_MS = params.getDeltaD() + params.getDeltaH();
		long limit = params.getRingLength() - params.getHMax() - params.getDeltaD();
		
		// Find the optimal block size.
		int bestBlockSize = 0;
		double bestBlockSizeTime = 0; 
		
		for (int i = 1; GRANULARITY_MS * (Math.pow(2, i) - 1) <= limit; i++) {
			double time = (params.getDMax() + params.getHMax() + GRANULARITY_MS * (Math.pow(2, i) - 1) / 2.0) / (double)i;
			if (bestBlockSize == 0 || time < bestBlockSizeTime) {
				bestBlockSize = i;
				bestBlockSizeTime = time;
			} else break;
		}
		
		BLOCK_SIZE = bestBlockSize;
		queue = new ArrayBlockingQueue<Boolean>(BLOCK_SIZE);
		Log.i("USER", "Block size: " + BLOCK_SIZE);
	}
			
	@Override
	public void send(boolean b, boolean last) throws InterruptedException {
		Log.i("USER", "Sender: Queuing bit " + (b ? "1" : "0"));
		queue.offer(b);
		
		// If the queue is not full, just add a bit.
		if (queue.size() < BLOCK_SIZE && !last) {
			return;
		}
		
		// Pad the end of the last block.
		while (queue.size() < BLOCK_SIZE) {
			queue.offer(false);
		}

		// Compute the ringing time.
		elapsedTime = 0;
		for (int i = 0; i < BLOCK_SIZE; i++)
			if (queue.poll())
				elapsedTime |= 1 << i;
		
		elapsedTime *= GRANULARITY_MS;
		Log.i("USER", "Sender: Expected ring time " + elapsedTime);
		
		// Ring.
		queue.clear();
		ring();
		
		this.mutexUp();
		
		// Allow h_{max} time for hangup.
		Thread.sleep(params.getHMax());
	}
	
	@Override
	public void callingCallback() {
		Timer timer = new Timer();
		
		// Allow d_{max} time for dialing, then wait for the computed time.
		timer.schedule(new TimerTask() {
			@Override
			public void run() {
				hangup();
				Log.i("USER", "Sender: Hangup");
			}
		}, params.getDMax() + elapsedTime);
		
		elapsedTime = this.getElapsedTimeMillis();
	}

	@Override
	public void callingHangupCallback() {
		// Make sure that the calling intent is off.
		hangup();
		
		Log.i("USER", "Sender: Call ended after " + (this.getElapsedTimeMillis() - elapsedTime));
		this.mutexDown();
	}
		
	@Override
	public void receivingCallback(String number) {
		// Start timing.
		elapsedTime = getElapsedTimeMillis();
	}

	@Override
	public void receivingHangupCallback() {
		// Finish timing.
		elapsedTime = getElapsedTimeMillis() - elapsedTime;
		this.mutexDown();
	}

	@Override
	public boolean receive() throws InterruptedException {
		if (queue.isEmpty()) {
			// Wait for the ringing.
			Log.i("USER", "Recvr : Waiting");
			this.mutexUp();
			Log.i("USER", "Recvr : Call ended after " + elapsedTime + " ms");
		
			int number = (int)((elapsedTime - params.getHMin()) / GRANULARITY_MS);

			// Deconstruct the received value.
			queue.clear();
			for (int i = 0; i < BLOCK_SIZE; i++)
				queue.offer((number & (1 << i)) != 0);
					
			
			Log.i("USER", "Recvr : Received as " + number);					
		}

		Log.i("USER", "Recvr : Received bit " + (queue.peek() ? "1" : "0"));
		return queue.poll();
	}
}
