package imp.ringing;

import android.content.Context;
import android.content.Intent;
import android.net.Uri;
import android.telephony.TelephonyManager;
import android.util.Log;

import com.android.internal.telephony.ITelephony;

import java.lang.reflect.Method;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

enum ProtocolState {Inactive, Sending, Receiving};
enum ProtocolCallbackType {Calling, Receiving, Hangup};

/**
 * An abstract superclass to all protocols. Defines basic routines used by all protocols.
 */
public abstract class Protocol {
	private CountDownLatch latch;
	
	protected String dialedNumber;
	protected ProtocolState protocolState = ProtocolState.Inactive;
	protected Parametrization params;
	
	public abstract void send(boolean b, boolean last) throws InterruptedException;		
	public abstract boolean receive() throws InterruptedException;
	
	/**
	 * Initializes the protocol with given parameters.
	 * @param params Parameters.
	 */
	public Protocol(Parametrization params) {
		this.params = params;
	};

	/**
	 * Handles the callback from CallController .
	 * @param type Callback type.
	 * @param number Caller phone number (in Receiving mode).
	 */
	public void phoneStateCallback(ProtocolCallbackType type, String number) {
		if (protocolState == ProtocolState.Inactive) return;
		
		switch (type) {
			case Calling:
				Log.i("USER", "Sender: Calling callback");
				callingCallback();
			break;
		
			case Receiving:
				Log.i("USER", "Recvr : Receiving callback");
				receivingCallback(number);
			break;
			
			case Hangup:
				if (protocolState == ProtocolState.Sending) {
					Log.i("USER", "Sender: Hangup callback");
					callingHangupCallback();
				} else {
					Log.i("USER", "Recvr : Hangup callback");
					receivingHangupCallback();
				}
			break;
		}
	}
	
	/** Called during Idle -> Offhook transition in sending mode. */  
	public abstract void callingCallback();
	/** Called during Idle -> Ringing transition in receiving mode. */  	
	public abstract void receivingCallback(String number);
	/** Called during Offhook -> Idle transition in sending mode. */
	public abstract void callingHangupCallback();
	/** Called during Ringing -> Idle transition in receiving mode. */
	public abstract void receivingHangupCallback();
	
	/**
	 * Sets the protocol state.
	 * @param state State.
	 */
	public void setProtocolState(ProtocolState state) {
		this.protocolState = state;
	}

	/**
	 * Sets the number to be dialed.
	 * @param number The called phone number.
	 */
	public void setDialedNumber(String number) {
		this.dialedNumber = number;
	}
	
	/**
	 * @return Telephony service object.
	 */
	protected ITelephony getTelephonyService() {
		TelephonyManager telephonyManager = (TelephonyManager)Main.context.getSystemService(Context.TELEPHONY_SERVICE);
		ITelephony telephonyService = null;
		
		try {
		    Log.i("DEBUG", "Get getTeleService...");
		    Class<TelephonyManager> c = TelephonyManager.class;
		    Method method = c.getDeclaredMethod("getITelephony");
		    method.setAccessible(true);
		    telephonyService = (ITelephony) method.invoke(telephonyManager);
		} catch (Exception e) {
		    e.printStackTrace();
		    Log.i("DEBUG",
		            "FATAL ERROR: could not connect to telephony subsystem");
		    Log.i("DEBUG", "Exception object: " + e);
		}
		
	    return telephonyService;
	}
	
	/** Dials the preset telephone number. */
	protected void ring() {
		try {
			Log.i("USER", "Sender: Ringing " + Uri.parse("tel:" + this.dialedNumber).toString());

			Intent intent = new Intent(Intent.ACTION_CALL);
			intent.setData(Uri.parse("tel:" + this.dialedNumber));
			Main.context.startActivity(intent);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/** Hangs up. */
	protected void hangup() {
		try {
			Log.i("USER", "Sender: Hanging up");
			this.getTelephonyService().endCall();
			
			Intent intent = new Intent(Main.context, Main.class);
			intent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
			Main.context.getApplicationContext().startActivity(intent);
		} catch (Exception e) {
			e.printStackTrace();
		}		
	}
	
	/**
	 * Gets hold of this protocol's common mutex.
	 * @throws InterruptedException
	 */
	protected void mutexUp() throws InterruptedException {
		latch = new CountDownLatch(1);
		latch.await();
	}

	/** 
	 * Gets hold of this protocol's common mutex for a limited time.
	 * @param milliseconds Number of miliseconds after which the mutex is released.
	 * @throws InterruptedException
	 */
	protected void mutexUpWithTimer(long milliseconds) throws InterruptedException {
		latch = new CountDownLatch(1);	
		latch.await(milliseconds, TimeUnit.MILLISECONDS);
	}

	/**
	 * Releases this protocol's common mutex.
	 */
	protected void mutexDown() {
		latch.countDown();
	}	
	
	/**
	 * @return The system time in milliseconds.
	 */
	protected long getElapsedTimeMillis() {
		return System.nanoTime() / /* ns to ms */ 1000000;
	}
}
