package imp.ringing;

import android.util.Log;

/**
 * A protocol which encodes 1 and 0 into ringing and silence, respectively. 
 */
public class RingSilenceProtocol extends Protocol {
	private long GRANULARITY_MS;
	private long GREATER_GRANULARITY_MS;
	private boolean isDSpanLarger = false;
	private boolean wasRinging = false;
	private boolean sendingStarted = false;
	private final long maxConsecutiveOnes;
	private long consecutiveOnes = 0;
	private long lastFrameTimeShift;

	/**
	 * Initializes this protocol.
	 * @param params Parameters.
	 */
	public RingSilenceProtocol(Parametrization params) {
		super(params);
		lastFrameTimeShift = 0;
		isDSpanLarger = (params.getDeltaD() >= params.getDeltaH());
		GRANULARITY_MS = (params.getDMax() > params.getHMax()) ? params.getDMax() : params.getHMax();
		GREATER_GRANULARITY_MS = GRANULARITY_MS + (isDSpanLarger ? params.getDeltaH() : params.getDeltaD());
		maxConsecutiveOnes = 1 + (long)(Math.floor(params.getRingLength() - GREATER_GRANULARITY_MS + params.getDMin()) / GREATER_GRANULARITY_MS);
		Log.i("USER", "Maximum consecutive ones: " + maxConsecutiveOnes);
	}
	
	@Override
	public void send(boolean b, boolean last) throws InterruptedException {
		// Synchronize according to d or h, whichever has smaller span.
		if (!sendingStarted) {
			sendingStarted = true;
			Log.i("USER", "Sender: Synchronizing dial");
			send(true, false);
			Log.i("USER", "Sender: Synchronizing hangup");
			send(false, false);
			
			GRANULARITY_MS = GREATER_GRANULARITY_MS;
		}
		
		// On maximum consecutive zeroes limit, send extra 0.
		if (consecutiveOnes == maxConsecutiveOnes) {
			Log.i("USER", "Sender : Sending extra 0");
			consecutiveOnes = 0;
			send(false, false);
		}
		
		// Accumulate the number of consecutive 1s.
		if (sendingStarted) {
			if (b) {
				++consecutiveOnes;
			} else {
				consecutiveOnes = 0;
			}
		}
			
		// Dial if a bit changes from 0 to 1.
		if (!wasRinging && b) {
			ring();
		}
		
		// Hang up if a bit changes from 1 to 0.
		if (wasRinging && !b) {
			hangup();
		}

		// Let the next frame pass.
		long realFrameTime = getElapsedTimeMillis();
		Thread.sleep(GRANULARITY_MS + lastFrameTimeShift);
		realFrameTime = getElapsedTimeMillis() - realFrameTime;
		lastFrameTimeShift = GRANULARITY_MS + lastFrameTimeShift - realFrameTime;
		Log.i("USER", "Sender: Frame shift " + lastFrameTimeShift + "ms");
		
		// Make sure that ringing stops after the transmission.
		if (last && wasRinging) {
			Log.i("USER", "Sender: Finished");
			hangup();
		}
	}

	@Override
	public void callingCallback() {
		wasRinging = true;
	}
	
	@Override
	public void callingHangupCallback() {
		// Make sure that the calling intent is off.
		hangup();
		
		wasRinging = false;
	}
	
	@Override
	public void receivingCallback(String number) {
		// Synchronize by d, if Delta d <= Delta h.
		if (!sendingStarted && !isDSpanLarger) {
			this.mutexDown();
		}
		Log.i("USER", "Recvr : Started ringing");
		wasRinging = true;
	}

	@Override
	public void receivingHangupCallback() {
		// Synchronize by h, if Delta h < Delta d.
		if (!sendingStarted && isDSpanLarger) {
			this.mutexDown();
		}
		
		Log.i("USER", "Recvr : Hung up");
		wasRinging = false;
	}

	@Override
	public boolean receive() throws InterruptedException {
		// Wait for the synchronization.
		if (!sendingStarted) {
			this.mutexUp();
			if (isDSpanLarger) {
				Log.i("USER", "Recvr: Calibrating by hangup");
				Thread.sleep(GRANULARITY_MS - params.getHMax());
			} else {
				Log.i("USER", "Recvr: Calibrating by ringing");
				Thread.sleep(GRANULARITY_MS);
			}
			
			GRANULARITY_MS = GREATER_GRANULARITY_MS;
			sendingStarted = true;
			consecutiveOnes = 0;
		}
		
		// Receive a bit and possibly an extra 0 after the maximum
		// number of consecutive 1s.
		while (true) {
			long realFrameTime = getElapsedTimeMillis();
			Thread.sleep(GRANULARITY_MS + lastFrameTimeShift);
			realFrameTime = getElapsedTimeMillis() - realFrameTime;
			lastFrameTimeShift = GRANULARITY_MS + lastFrameTimeShift - realFrameTime;
			Log.i("USER", "Recvr : Frame shift " + lastFrameTimeShift + "ms");
			
			if (consecutiveOnes == maxConsecutiveOnes) {
				Log.i("USER", "Recvr : Ignoring extra 0");
				consecutiveOnes = 0;
				continue;
			}

			if (wasRinging) {
				++consecutiveOnes;
			} else {				
				consecutiveOnes = 0;
			}
			
			break;
		}
		
		Log.i("USER", "Recvr : Answer");
		return wasRinging;
	}
	
}
